/*-------------------------------------------------------------------------
 * pgsnmpd.c
 *
 *      Global definitions for the main PostgreSQL SNMP Daemon.
 *
 *      Copyright (c) 2004-2005, PostgreSQL Global Development Group
 *      Author: Hiroshi Saito
 *
 *      $Id: pgsnmpd.c,v 1.00 2005/11/16 13:14:05 saito Exp $
 *-------------------------------------------------------------------------
 */

#include <unistd.h>

#include "postgres_fe.h"
#include "libpq-fe.h"

#ifdef PACKAGE_BUGREPORT
#undef PACKAGE_BUGREPORT
#endif
#ifdef PACKAGE_NAME
#undef PACKAGE_NAME
#endif
#ifdef PACKAGE_STRING
#undef PACKAGE_STRING
#endif
#ifdef PACKAGE_TARNAME
#undef PACKAGE_TARNAME
#endif
#ifdef PACKAGE_VERSION
#undef PACKAGE_VERSION
#endif

#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include <signal.h>
#include <ctype.h>

/*
 * define strings
 */

#define PGSNMP_DAEMON "PGSQL-SNMP-0.1b1"

/* This can be used henceforth Version 8.1
#define PG_VERSION_STR "PostgreSQL 8.1.0 (release)"
*/
#define SUB_AGENT	1

char *pgsnmp_version = PG_VERSION_STR;

/* MIB object enterprises is private */
#define enterprises	1, 3, 6, 1, 4, 1
/* rfc1697 RDBMS Objects  (not use)
#define rdbmsoid	1, 3, 6, 1, 2, 1, 39
*/

/*
 * PostgreSQL Connection
 */
PGconn     *dbconn = NULL;
char       *conninfo = {"dbname=template1"};

static int keep_running = 0;
struct timeval snmp_boottime;

oid  PostgreSQLMIB_pgsqlSrvInfo[] = { enterprises, 5432, 1, 4, 1 };
   //    { rdbmsoid, 1, 4, 1, 1 };
oid  PostgreSQLMIB_pgsqlSrvTable[] = { enterprises, 5432, 1, 4, 2 };
   //    { rdbmsoid, 1, 4, 1, 2 };

netsnmp_table_data_set *pgsqlSrvInfo = NULL;
netsnmp_table_data_set *pgsqlSrvTable = NULL;

/* 
 * Statistics SQL 
 */
static char *SQL_getversion = {
        "SELECT version();"
};
static char *SQL_getuptime = {
	"SELECT CASE WHEN usesuper THEN pg_postmaster_start_time() "
	"ELSE NULL END as upsince FROM pg_user WHERE usename=current_user;"
};
static char *SQL_getdatabase = {
	"SELECT d.datname, r.rolname,"
	"pg_catalog.pg_encoding_to_char(d.encoding),"
	"s.numbackends, s.xact_commit, s.xact_rollback,"
	"s.blks_read, s.blks_hit,"
	"pg_size_pretty(pg_database_size(s.datid)) as size "
	"FROM pg_catalog.pg_database d "
	"LEFT JOIN pg_catalog.pg_roles r ON d.datdba = r.oid "
	"LEFT JOIN pg_stat_database s ON d.datname = s.datname "
	"ORDER BY d.datname;"
};


/*
 * database Infomation
 */

/* MIB = 1, 3, 6, 5432, 1, 4, 1 */
static char *pgsqlVersion = NULL;
static char *pgsqlUptime = NULL;
static int   pgsqlState = 0;

/* MIB = 1, 3, 6, 5432, 1, 4, 2  *N record */
typedef struct {
	char *pgsqlDataname;
	char *pgsqlOwner;
	char *pgsqlEncoding;
	int   pgsqlDumbackends;
	int   pgsqlXact_commit;
	int   pgsqlXact_rollback;
	int   pgsqlBlks_read;
	int   pgsqlBlks_hit;
	char *pgsqlKBsize;
} pgsqlDatabase_Variable;

static int pgsqlDatabase_cnt = 0;
static pgsqlDatabase_Variable *pgsqlDatabase[] = {NULL};

static PGconn
*db_connect(char *constring)
{
	PGconn     *dbconn;
	dbconn = PQconnectdb(constring);
	if (dbconn == NULL)
	{
		return NULL;
	}
	if (PQstatus(dbconn) != CONNECTION_OK)
	{
		snmp_log(LOG_ERR,"%s\n",PQerrorMessage(dbconn));
		PQfinish(dbconn);
		return NULL;
	}

	return dbconn;
}	

static
bool IsAlive()
{
    if (!pgsqlState)
    {
      if ((dbconn = db_connect(conninfo)))
        pgsqlState = 1;
      else
        return false;
    }

    PGresult *qryRes = PQexec(dbconn, "SELECT 1;");
    if (PQresultStatus(qryRes) != PGRES_TUPLES_OK)
    {
        PQclear(qryRes);
        PQfinish(dbconn);
        pgsqlState = 0;
        dbconn = NULL;
        return false;
    }

    return true;
}

static
void pq_connectinfo()
{
    if (!IsAlive())
    {
	pgsqlVersion = PGSNMP_DAEMON;
	pgsqlUptime = NULL;
        return;
    }

    PGresult *qryRes = PQexec(dbconn, SQL_getversion);
    if (PQresultStatus(qryRes) != PGRES_TUPLES_OK)
    {
        PQclear(qryRes);
	pgsqlVersion = PGSNMP_DAEMON;
        return;
    }

    pgsqlVersion = strdup(PQgetvalue(qryRes, 0, 0));
    PQclear(qryRes);
#ifdef DEBUG
    snmp_log(LOG_INFO,"%s\n", pgsqlVersion);
#endif

    qryRes = PQexec(dbconn, SQL_getuptime);
    if (PQresultStatus(qryRes) != PGRES_TUPLES_OK)
    {
        PQclear(qryRes);
	pgsqlUptime = NULL;
        return;
    }

    pgsqlUptime = strdup(PQgetvalue(qryRes, 0, 0));
    PQclear(qryRes);
#ifdef DEBUG
    snmp_log(LOG_INFO,"%s\n", pgsqlUptime);
#endif

}

static
void pq_database()
{
    int res_cnt;

    if (IsAlive())
    {
	for(res_cnt = pgsqlDatabase_cnt; res_cnt > 0 ;res_cnt--)
	{
	    free(pgsqlDatabase[res_cnt-1]->pgsqlDataname);
	    free(pgsqlDatabase[res_cnt-1]->pgsqlOwner);
	    free(pgsqlDatabase[res_cnt-1]->pgsqlEncoding);
	    free(pgsqlDatabase[res_cnt-1]->pgsqlKBsize);
	    free(pgsqlDatabase[res_cnt-1]);
	    pgsqlDatabase[res_cnt-1] = NULL;
	}
	pgsqlDatabase_cnt = 0;
    }
    else
	return;

    PGresult *qryRes = PQexec(dbconn, SQL_getdatabase);
    if (PQresultStatus(qryRes) != PGRES_TUPLES_OK)
    {
        PQclear(qryRes);
	pgsqlVersion = PGSNMP_DAEMON;
        return;
    }

    pgsqlDatabase_cnt = PQntuples(qryRes);
#ifdef DEBUG
    snmp_log(LOG_INFO,"CNT=%d\n", pgsqlDatabase_cnt);
#endif
    for( res_cnt = 0; res_cnt < pgsqlDatabase_cnt; res_cnt++)
    {
	pgsqlDatabase[res_cnt] = (pgsqlDatabase_Variable *)malloc(sizeof(pgsqlDatabase_Variable));
	pgsqlDatabase[res_cnt]->pgsqlDataname = strdup(PQgetvalue(qryRes, res_cnt, 0));
	pgsqlDatabase[res_cnt]->pgsqlOwner = strdup(PQgetvalue(qryRes, res_cnt, 1));
	pgsqlDatabase[res_cnt]->pgsqlEncoding = strdup(PQgetvalue(qryRes, res_cnt, 2));
	pgsqlDatabase[res_cnt]->pgsqlDumbackends = atoi(PQgetvalue(qryRes, res_cnt, 3));
	pgsqlDatabase[res_cnt]->pgsqlXact_commit = atoi(PQgetvalue(qryRes, res_cnt, 4));
	pgsqlDatabase[res_cnt]->pgsqlXact_rollback = atoi(PQgetvalue(qryRes, res_cnt, 5));
	pgsqlDatabase[res_cnt]->pgsqlBlks_read = atoi(PQgetvalue(qryRes, res_cnt, 6));
	pgsqlDatabase[res_cnt]->pgsqlBlks_hit = atoi(PQgetvalue(qryRes, res_cnt, 7));
	pgsqlDatabase[res_cnt]->pgsqlKBsize = strdup(PQgetvalue(qryRes, res_cnt, 8));
#ifdef DEBUG
        snmp_log(LOG_INFO,"%s-%s-%s-%d-%d-%d-%d-%d-%s\n",
		pgsqlDatabase[res_cnt]->pgsqlDataname,
		pgsqlDatabase[res_cnt]->pgsqlOwner,
		pgsqlDatabase[res_cnt]->pgsqlEncoding,
		pgsqlDatabase[res_cnt]->pgsqlDumbackends,
		pgsqlDatabase[res_cnt]->pgsqlXact_commit,
		pgsqlDatabase[res_cnt]->pgsqlXact_rollback,
		pgsqlDatabase[res_cnt]->pgsqlBlks_read,
		pgsqlDatabase[res_cnt]->pgsqlBlks_hit,
		pgsqlDatabase[res_cnt]->pgsqlKBsize);
#endif

    }

    PQclear(qryRes);
	
}


static void
stop_server(int signo) 
{
	pgsqlState = keep_running = 0;
	snmp_log(LOG_ERR, "\nExiting on signal %d\n", signo);
}

static int
pgSnmpRequest_pgsqlSrvInfo(netsnmp_mib_handler *handler,
	netsnmp_handler_registration *reginfo,
	netsnmp_agent_request_info *reqinfo,
	netsnmp_request_info *requests)
{
	netsnmp_request_info *request;
	netsnmp_table_request_info *table_info;
	netsnmp_table_row *table_row;
    
	/* retry infomation */
  	pq_connectinfo();
  	pq_database();

        for (request = requests; request; request = request->next)
        {
		table_info = netsnmp_extract_table_info(request);
		netsnmp_variable_list *var = request->requestvb;
		table_row = netsnmp_extract_table_row(request);

                switch (reqinfo->mode)
                {
                case MODE_GETNEXT:
                case MODE_GET:
#ifdef DEBUG
  			snmp_log(LOG_INFO,"MODE_GET(%d:%d-OID:%d-%d-%d(%d-%d)).\n",
			table_info->colnum,
			table_info->number_indexes,
			table_info->index_oid,
			table_info->index_oid_len,
			table_info->reg_info->number_indexes,
			table_info->reg_info->min_column,
			table_info->reg_info->max_column);
#endif

			if (table_row)
			{
			  switch (table_info->colnum) {
			  case 1:
				snmp_set_var_typed_value(var, ASN_OCTET_STR,
                             		pgsqlVersion,strlen(pgsqlVersion));
				break;

			  case 2:
    				if (pgsqlUptime)
    				{
				   snmp_set_var_typed_value(var, ASN_OCTET_STR,
                           		pgsqlUptime, strlen(pgsqlUptime));
    				}
				else
    				{
				   snmp_set_var_typed_value(var, ASN_OCTET_STR,
                           		"", strlen(""));
    				}
				
				break;

			  case 3:
				snmp_set_var_typed_value(var, ASN_INTEGER,
                 			(char*)&pgsqlState,sizeof(pgsqlState));
				break;
                	  default:
                        	break;
			  }
			}

			/*
			netsnmp_table_data_delete_row(table_row);
			*/
                default:
  			snmp_log(LOG_INFO,"MODE_OTHERS.\n");
                        break;
                }
        }                                                                                                 
        return SNMP_ERR_NOERROR;
}                                   

static int
pgSnmpRequest_pgsqlSrvTable(netsnmp_mib_handler *handler,
	netsnmp_handler_registration *reginfo,
	netsnmp_agent_request_info *reqinfo,
	netsnmp_request_info *requests)
{
	netsnmp_request_info *request;
	netsnmp_table_request_info *table_info;
	netsnmp_table_row *table_row;

	subtree_context_cache *context_ptr;

/* ************** not made here yet. ******* */
        return SNMP_ERR_NOERROR;
/* ************** not made here yet. ******* */

#ifdef DEBUG
  	snmp_log(LOG_INFO,"HAND(%d-%d-%d)\n",reginfo->rootoid,
		reginfo->range_subid,reginfo->global_cacheid);
#endif    
	/* retry infomation */
  	pq_connectinfo();
  	pq_database();
	
        for (request = requests; request; request = request->next )
        {
		table_info = netsnmp_extract_table_info(request);
		table_row = netsnmp_extract_table_row(request);

		netsnmp_variable_list *var = request->requestvb;

		int tablenumber = table_info->number_indexes;
		
		context_ptr = (subtree_context_cache *)
            		netsnmp_extract_iterator_context(requests);

                switch (reqinfo->mode)
                {
                case MODE_GETNEXT:
                case MODE_GET:
#ifdef DEBUG
  			snmp_log(LOG_INFO,"MODE_GET(%d:%d-OID:%d-%d-%d(%d-%d)).\n",
			table_info->colnum,
			table_info->number_indexes,
			table_info->index_oid,
			table_info->index_oid_len,
			table_info->reg_info->number_indexes,
			table_info->reg_info->min_column,
			table_info->reg_info->max_column);
#endif

			if (table_row)
			{
			  switch (table_info->colnum) {
			  case 1:
				if (( tablenumber < pgsqlDatabase_cnt ) &&
				    ( tablenumber >= 0))
				snmp_set_var_typed_value(var, ASN_OCTET_STR,
                             	pgsqlDatabase[tablenumber]->pgsqlDataname,
                     		strlen(pgsqlDatabase[tablenumber]->pgsqlDataname));
#if 1
  			snmp_log(LOG_INFO,"DEBUG(%s-%d).\n",
				pgsqlDatabase[tablenumber]->pgsqlDataname,
					tablenumber);
#endif
                		break;
			  case 2:
				if (( tablenumber < pgsqlDatabase_cnt ) &&
				    ( tablenumber >= 0))
				snmp_set_var_typed_value(var, ASN_OCTET_STR,
                             	pgsqlDatabase[tablenumber]->pgsqlOwner,
                     		strlen(pgsqlDatabase[tablenumber]->pgsqlOwner));
                		break;
			  case 3:
				if (( tablenumber < pgsqlDatabase_cnt ) &&
				    ( tablenumber >= 0))
				snmp_set_var_typed_value(var, ASN_OCTET_STR,
                             	pgsqlDatabase[tablenumber]->pgsqlEncoding,
                     		strlen(pgsqlDatabase[tablenumber]->pgsqlEncoding));
                		break;
			  case 9:
				if (( tablenumber < pgsqlDatabase_cnt ) &&
				    ( tablenumber >= 0))
				snmp_set_var_typed_value(var, ASN_OCTET_STR,
                             	pgsqlDatabase[tablenumber]->pgsqlKBsize,
                     		strlen(pgsqlDatabase[tablenumber]->pgsqlKBsize));
                		break;
            		  default:
                		netsnmp_set_request_error(reqinfo, requests,
                                          SNMP_NOSUCHOBJECT);
                		break;
            		  }
            		}

                        break;

                default:
  			snmp_log(LOG_INFO,"MODE_OTHERS.\n");
                        break;
                }
        }                                                                                                 
        return SNMP_ERR_NOERROR;
}                                   

/* Struct PostgreSQL MIB Data */
static
void PostgreSQLMIBset(void)
{
	int tablenumber;

	netsnmp_table_row *row;


	/*
	 * register the table set pgsqlSrvInfo
	 */

	pgsqlSrvInfo = netsnmp_create_table_data_set("pgsqlSrvInfo");
	pgsqlSrvInfo->allow_creation = 1;

	netsnmp_table_dataset_add_index(pgsqlSrvInfo, ASN_INTEGER);
	netsnmp_table_set_multi_add_default_row(pgsqlSrvInfo,
                                            1, ASN_OCTET_STR, 1, NULL, 0,
                                            2, ASN_OCTET_STR, 1, NULL, 0,
                                            3, ASN_INTEGER, 1, NULL, 0,
                                            0 /* done */ );
	netsnmp_register_table_data_set(netsnmp_create_handler_registration
    					("pgsqlSrvInfo", 
					pgSnmpRequest_pgsqlSrvInfo,
    					PostgreSQLMIB_pgsqlSrvInfo,
   					OID_LENGTH(PostgreSQLMIB_pgsqlSrvInfo),
   					HANDLER_CAN_RWRITE), pgsqlSrvInfo,
					NULL);

	row = netsnmp_create_table_data_row();
	tablenumber = 1;
	netsnmp_table_row_add_index(row, ASN_INTEGER, &tablenumber,
						 sizeof(tablenumber));

	netsnmp_set_row_column(row, 1, ASN_OCTET_STR,
                           pgsqlVersion, strlen(pgsqlVersion));
	netsnmp_mark_row_column_writable(row, 1, 1);

	if (pgsqlUptime)
	{
        	netsnmp_set_row_column(row, 2, ASN_OCTET_STR,
                           pgsqlUptime, strlen(pgsqlUptime));
        	netsnmp_mark_row_column_writable(row, 2, 1);
    	}

       	netsnmp_set_row_column(row, 3, ASN_INTEGER,
                 (char*)&pgsqlState,sizeof(pgsqlState));
       	netsnmp_mark_row_column_writable(row, 3, 1);
       

	netsnmp_table_dataset_add_row(pgsqlSrvInfo, row);
	netsnmp_register_auto_data_table(pgsqlSrvInfo, NULL);

	/*
	 * register the table set pgsqlSrvTable
	 */

	pgsqlSrvTable = netsnmp_create_table_data_set("pgsqlSrvTable");
	pgsqlSrvTable->allow_creation = 1;

	netsnmp_table_dataset_add_index(pgsqlSrvTable, ASN_INTEGER);
	netsnmp_table_set_multi_add_default_row(pgsqlSrvTable,
                                            1, ASN_OCTET_STR, 1, NULL, 0,
                                            2, ASN_OCTET_STR, 1, NULL, 0,
                                            3, ASN_OCTET_STR, 1, NULL, 0,
                                            4, ASN_INTEGER, 1, NULL, 0,
                                            5, ASN_INTEGER, 1, NULL, 0,
                                            6, ASN_INTEGER, 1, NULL, 0,
                                            7, ASN_INTEGER, 1, NULL, 0,
                                            8, ASN_INTEGER, 1, NULL, 0,
                                            9, ASN_OCTET_STR, 1, NULL, 0,
                                            0 /* done */ );
	netsnmp_register_table_data_set(netsnmp_create_handler_registration
    					("pgsqlSrvTable", 
					pgSnmpRequest_pgsqlSrvTable,
    					PostgreSQLMIB_pgsqlSrvTable,
   					OID_LENGTH(PostgreSQLMIB_pgsqlSrvTable),
   					HANDLER_CAN_RWRITE), pgsqlSrvTable,
					NULL);

	for(tablenumber = 1; tablenumber <= pgsqlDatabase_cnt; tablenumber++)
	{

       		row = netsnmp_create_table_data_row();
       		netsnmp_table_row_add_index(row, ASN_INTEGER, &tablenumber,
						 sizeof(tablenumber));

       		netsnmp_set_row_column(row, 1, ASN_OCTET_STR,
                     pgsqlDatabase[tablenumber-1]->pgsqlDataname, 
		     strlen(pgsqlDatabase[tablenumber-1]->pgsqlDataname));
       		netsnmp_mark_row_column_writable(row, 1, 1);
       
       		netsnmp_set_row_column(row, 2, ASN_OCTET_STR,
                     pgsqlDatabase[tablenumber-1]->pgsqlOwner, 
		     strlen(pgsqlDatabase[tablenumber-1]->pgsqlOwner));
       		netsnmp_mark_row_column_writable(row, 2, 1);
       
       		netsnmp_set_row_column(row, 3, ASN_OCTET_STR,
                     pgsqlDatabase[tablenumber-1]->pgsqlEncoding, 
		     strlen(pgsqlDatabase[tablenumber-1]->pgsqlEncoding));
       		netsnmp_mark_row_column_writable(row, 3, 1);

       		netsnmp_set_row_column(row, 4, ASN_INTEGER,
                     (char*)&pgsqlDatabase[tablenumber-1]->pgsqlDumbackends,
		     sizeof(pgsqlDatabase[tablenumber-1]->pgsqlDumbackends));
       		netsnmp_mark_row_column_writable(row, 4, 1);

       		netsnmp_set_row_column(row, 5, ASN_INTEGER,
                     (char*)&pgsqlDatabase[tablenumber-1]->pgsqlXact_commit,
		     sizeof(pgsqlDatabase[tablenumber-1]->pgsqlXact_commit));
       		netsnmp_mark_row_column_writable(row, 5, 1);
       
       		netsnmp_set_row_column(row, 6, ASN_INTEGER,
                     (char*)&pgsqlDatabase[tablenumber-1]->pgsqlXact_rollback,
		     sizeof(pgsqlDatabase[tablenumber-1]->pgsqlXact_rollback));
       		netsnmp_mark_row_column_writable(row, 6, 1);
       
       		netsnmp_set_row_column(row, 7, ASN_INTEGER,
                     (char*)&pgsqlDatabase[tablenumber-1]->pgsqlBlks_read,
		     sizeof(pgsqlDatabase[tablenumber-1]->pgsqlBlks_read));
       		netsnmp_mark_row_column_writable(row, 7, 1);

       		netsnmp_set_row_column(row, 8, ASN_INTEGER,
                     (char*)&pgsqlDatabase[tablenumber-1]->pgsqlBlks_hit,
		     sizeof(pgsqlDatabase[tablenumber-1]->pgsqlBlks_hit));
       		netsnmp_mark_row_column_writable(row, 8, 1);

       		netsnmp_set_row_column(row, 9, ASN_OCTET_STR,
                     pgsqlDatabase[tablenumber-1]->pgsqlKBsize,
		     strlen(pgsqlDatabase[tablenumber-1]->pgsqlKBsize));
       		netsnmp_mark_row_column_writable(row, 9, 1);

      		netsnmp_table_dataset_add_row(pgsqlSrvTable, row);
	}

	netsnmp_register_auto_data_table(pgsqlSrvTable, NULL);

}

static void
usage(void)
{
	fprintf(stderr, "Versin %s\n",PGSNMP_DAEMON);
	fprintf(stderr, "usage: pgsnmpd [-s] [-b] [-c FILE ] [-x address ] [-g] [-C \"Connect String\"]\n");
	fprintf(stderr, "   -s : SNMP subagent running\n");
	fprintf(stderr, "   -b : run in the background\n");
	fprintf(stderr, "   -c : configuration file name\n");
	fprintf(stderr, "   -g : use syslog\n");
	fprintf(stderr, "   -C : libpq connect strings\n");
	fprintf(stderr, "   -x : address:port of a network interface.\n");
	fprintf(stderr, "   -V : display version strings\n");
}

int
main(int argc, char *argv[]) 
{
	int agentx_subagent=0;	/* SNMP master agent */
	int background = 0;	/* run the background */
	int syslog = 0;		/* use syslog */

	int	arg;

#ifdef DEBUG
	char *strv;
#endif

	char	options[128] = "sbc:gC:x:V-:";

	while ((arg = getopt(argc, argv, options)) != EOF)
	{
		switch (arg)
		{
		case 'V':
			usage();
			exit(0);
		case 'c':
			if (optarg != NULL) {
                		netsnmp_ds_set_string(NETSNMP_DS_LIBRARY_ID,
                                      NETSNMP_DS_LIB_OPTIONALCONFIG, optarg);
            		} else {
                		usage();
				exit(1);
            		}
            		break;
        	case 'g':
            		syslog = 1;
            		break;
        	case 'b':
            		background = 1;
            		break;
        	case 's':
            		agentx_subagent = 1;
            		break;
        	case 'x':
            		if (optarg != NULL) {
                		netsnmp_ds_set_string(NETSNMP_DS_APPLICATION_ID,
                                      NETSNMP_DS_AGENT_X_SOCKET, optarg);
            		} else {
                		usage();
				exit(1);
            		}
            		break;
        	case 'C':
            		if (optarg != NULL) {
                		conninfo = optarg;
            		} else {
                		usage();
				exit(1);
            		}
            		break;
        	default:
                	usage();
			exit(1);
        	}
	}

  	gettimeofday(&snmp_boottime, (struct timezone *)0);

  	if (syslog)
    		snmp_enable_calllog();
  	else
    		snmp_enable_stderrlog();

  	if ((dbconn = db_connect(conninfo)))
     		pgsqlState = 1;

  	/* Initialise */
  	pq_connectinfo();
  	pq_database();

  	if (agentx_subagent)
  	{
    		/* make us a agentx client. */
    		netsnmp_ds_set_boolean(NETSNMP_DS_APPLICATION_ID,
                           NETSNMP_DS_AGENT_ROLE, SUB_AGENT);

/*
    		netsnmp_ds_set_boolean(NETSNMP_DS_APPLICATION_ID, 
			   NETSNMP_DS_AGENT_AGENTX_MASTER, 0);

*/
    		if (getuid() !=0)
      			netsnmp_ds_set_boolean(NETSNMP_DS_APPLICATION_ID, 
			     NETSNMP_DS_AGENT_NO_ROOT_ACCESS, 1);

#ifdef DEBUG
    		if (netsnmp_ds_get_boolean(NETSNMP_DS_APPLICATION_ID, 
                               NETSNMP_DS_AGENT_ROLE)) 
  			snmp_log(LOG_INFO,"SUB_AGENT running.\n");
#endif

	}
/*
  	netsnmp_ds_set_int(NETSNMP_DS_APPLICATION_ID,
                       NETSNMP_DS_AGENT_AGENTX_TIMEOUT, -1);
  	netsnmp_ds_set_int(NETSNMP_DS_APPLICATION_ID,
                       NETSNMP_DS_AGENT_AGENTX_RETRIES, -1);
*/

#if 0
  	netsnmp_ds_register_config(ASN_OCTET_STR, "snmp", "public",
		       NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_COMMUNITY);

#endif
#ifdef DEBUG
	if(netsnmp_ds_get_string(NETSNMP_DS_APPLICATION_ID, 
			NETSNMP_DS_AGENT_X_SOCKET))

		snmp_log(LOG_INFO,"AGENT_X_SOCKET:%s\n",
			netsnmp_ds_get_string(NETSNMP_DS_APPLICATION_ID, 
						NETSNMP_DS_AGENT_X_SOCKET));
 	if(netsnmp_ds_get_string(NETSNMP_DS_APPLICATION_ID, 
			NETSNMP_DS_AGENT_PORTS))
    		snmp_log(LOG_INFO,"AGENT_PORTS:%s\n",
			netsnmp_ds_get_string(NETSNMP_DS_APPLICATION_ID, 
						NETSNMP_DS_AGENT_PORTS));
#endif

	/* run in background */
	if (background && netsnmp_daemonize(1, !syslog))
		exit(1);

/* initialize tcpip */
SOCK_STARTUP;

	init_agent(PGSNMP_DAEMON);

	/* default MIB Data */
	PostgreSQLMIBset();

	init_snmp(PGSNMP_DAEMON);

	if (!agentx_subagent)
		init_master_agent();

	keep_running = 1;

	signal(SIGTERM, stop_server);
	signal(SIGINT, stop_server);

	snmp_log(LOG_INFO,"%s is up and running.\n", PGSNMP_DAEMON);

	while(keep_running)
	{

		agent_check_and_process(1);
#ifdef DEBUG
		strv = netsnmp_ds_get_string(NETSNMP_DS_LIBRARY_ID,
				NETSNMP_DS_LIB_COMMUNITY );
		if (strv)
			snmp_log(LOG_INFO, strv);
#endif

	}

	/* shutdown */
	if (!agentx_subagent)
		snmp_shutdown(PGSNMP_DAEMON);

SOCK_CLEANUP;

	return 0;

}
